#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <lil.h>

struct define {
    char* name;
    char* value;
    size_t vlen;
};

lil_t lil;
int embedded;
struct define* define;
size_t defines;
char* last_embed_str;

/* Under Windows we use WriteFile to avoid any CR/LF parsing from the C lib */
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
void win_write(const char* msg)
{
    static HANDLE hOut = INVALID_HANDLE_VALUE;
    DWORD towrite = strlen(msg), written = 0, justwritten;
    if (hOut == INVALID_HANDLE_VALUE) hOut = GetStdHandle(STD_OUTPUT_HANDLE);
    while (written < towrite) {
        if (!WriteFile(hOut, msg, towrite, &justwritten, NULL)) break;
        written += justwritten;
    }
}
#endif

void process_file(const char* path);

/* Defines */
struct define* find_define(const char* name)
{
    size_t i;
    for (i=0; i < defines; i++)
        if (!strcmp(name, define[i].name)) return define + i;
    return NULL;
}
void set_define(const char* name, const char* value)
{
    struct define* def = find_define(name);
    if (!def) {
        define = realloc(define, sizeof(struct define)*(defines + 1));
        def = define + defines++;
        def->name = strdup(name);
        def->value = NULL;
    };
    free(def->value);
    def->value = strdup(value);
    def->vlen = strlen(value);
}
const char* get_define(const char* name)
{
    struct define* def = find_define(name);
    return def ? def->value : "";
}

/* Load a single file */
char* load_file(const char* path)
{
    FILE* f = fopen(path, "rb");
    size_t size;
    char* buff;
    if (!f) return NULL;
    fseek(f, 0, SEEK_END);
    size = ftell(f);
    fseek(f, 0, SEEK_SET);
    buff = malloc(size + 1);
    if (fread(buff, 1, size, f) != size) {
        free(buff);
        fclose(f);
        return NULL;
    }
    buff[size] = '\0';
    fclose(f);
    return buff;
}

/* LIL callbacks filtering */
void cb_error(lil_t lil, size_t pos, const char* msg)
{
    fprintf(stderr, "lip: error at %u: %s\n", (unsigned)pos, msg);
}
const char* cb_embed_filter(lil_t lil, const char* msg)
{
    char *ret;
    char name[256];
    size_t i, starti, rsize, rlen, nlen;
    struct define* def;
    free(last_embed_str);
    last_embed_str = NULL;
    if (!strchr(msg, '$')) return msg;
    rsize = strlen(msg) + 16384;
    ret = malloc(rsize);
    rlen = 0;
    for (i=0; msg[i]; i++) {
        if (msg[i] == '$') {
            nlen = 0;
            starti = i;
            for (i++; msg[i] && msg[i] != '$' && !isspace(msg[i]); i++) {
                name[nlen++] = msg[i];
                if (nlen + 1 == sizeof(name)) break;
            }
            if (msg[i] == '$') {
                name[nlen] = 0;
                def = find_define(name);
            } else def = NULL;
            if (def) {
                if (rlen + def->vlen >= rsize) {
                    rsize += def->vlen + 16384;
                    ret = realloc(ret, rsize);
                }
                memcpy(ret + rlen, def->value, def->vlen);
                rlen += def->vlen;
            } else {
                if (rlen + (i - starti + 1) >= rsize) {
                    rsize += i - starti + 16384;
                    ret = realloc(ret, rsize);
                }
                memcpy(ret + rlen, msg + starti, i - starti + 1);
                rlen += i - starti + 1;
            }
        } else {
            if (rlen >= rsize) {
                rsize += 16384;
                ret = realloc(ret, rsize);
            }
            ret[rlen++] = msg[i];
        }
    }
    if (rlen >= rsize) {
        rsize += 1;
        ret = realloc(ret, rsize);
    }
    ret[rlen] = 0;
    return ret;
}

/* Register LIL functions */
lil_value_t fnc_file(lil_t lil, size_t argc, lil_value_t* argv)
{
    const char* path;
    if (argc < 1) return NULL;
    path = lil_to_string(argv[0]);
    embedded++;
    if (path) process_file(path);
    embedded--;
    return NULL;
}
lil_value_t fnc_system(lil_t lil, size_t argc, lil_value_t* argv)
{
    const char* cmd;
    if (argc < 1) return NULL;
    cmd = lil_to_string(argv[0]);
    if (cmd) system(cmd);
    return NULL;
}
lil_value_t fnc_define(lil_t lil, size_t argc, lil_value_t* argv)
{
    size_t i;
    for (i=0; i + 1 < argc; i += 2)
        set_define(lil_to_string(argv[i]), lil_to_string(argv[i + 1]));
    return NULL;
}
lil_value_t fnc_getdef(lil_t lil, size_t argc, lil_value_t* argv)
{
    struct define* def = argc > 0 ? find_define(lil_to_string(argv[0])) : NULL;
    return def && def->vlen ? lil_alloc_string(def->value) : NULL;
}
lil_value_t fnc_undef(lil_t lil, size_t argc, lil_value_t* argv)
{
    size_t i;
    for (i=0; i < argc; i++)
        set_define(lil_to_string(argv[i]), "");
    return NULL;
}
void register_functions(void)
{
    lil_register(lil, "file", fnc_file);
    lil_register(lil, "system", fnc_system);
    lil_register(lil, "define", fnc_define);
    lil_register(lil, "getdef", fnc_getdef);
    lil_register(lil, "undef", fnc_undef);
}

/* Process a single file */
void process_file(const char* path)
{
    /* Load the file */
    char* code = load_file(path);
    char* out;
    if (!code) {
        fprintf(stderr, "failed to load %s\n", path);
        return;
    }
    /* Process the file using LIL */
    out = lil_embedded(lil, code, LIL_EMBED_NOFLAGS);
    if (out) {
        if (out[0]) {
            if (embedded) lil_write(lil, out);
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
            else win_write(out);
#else
            else fwrite(out, 1, strlen(out), stdout);
#endif
        }
        lil_freemem(out);
    }
    /* Release */
    free(code);
}

int main(int argc, const char** argv)
{
    int i;
    
    /* Initialize LIL */
    lil = lil_new();
    lil_callback(lil, LIL_CALLBACK_ERROR, (lil_callback_proc_t)cb_error);
    lil_callback(lil, LIL_CALLBACK_EMBEDDEDFILTER, (lil_callback_proc_t)cb_embed_filter);
    register_functions();
    
    /* Parse parameters */
    for (i=1; i < argc; i++)
        if (argv[i][0] == '-') {
            if (!strcmp(argv[i], "--help")) {
                printf("LIL Preprocessor (lip) Help:\n\n");
                printf("Usage: lip [--code|-c code] [files]\n\n");
                printf("lip will process each one of the given files. Any\n");
                printf("definitions a file will still be valid in following\n");
                printf("files\n");
                printf("The -c or --code <code> parameter can also be used\n");
                printf("to execute LIL code directly.\n");
                return 0;
            } else if (!strcmp(argv[i], "-c") || !strcmp(argv[i], "--code")) {
                i++;
                if (i < argc)
                    lil_free_value(lil_parse(lil, argv[i], strlen(argv[i]), 0));
                continue;
            }
            fprintf(stderr, "unknown option %s\n", argv[i]);
            return 1;
        } else process_file(argv[i]);
    
    /* Shutdown LIL */
    lil_free(lil);

    /* Free any lingering memory */
    free(last_embed_str);
    last_embed_str = NULL;
    return 0;
}
